#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Mon Nov 14 09:08:25 2016
# Last Modified : ven. 30 mars 2018 17:28:23 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import curses
from cpyvke.utils.kernel import kernel_dic, start_new_kernel, \
    shutdown_kernel, connect_kernel
from cpyvke.utils.comm import send_msg
from cpyvke.objects.panel import ListPanel


class KernelWin(ListPanel):
    """ Kernel manager panel """

    def __init__(self, app, sock, logger):
        super(KernelWin, self).__init__(app, sock, logger)

    @property
    def panel_name(self):
        return 'kernel-manager'

    @property
    def title(self):
        return ' Kernel Manager '

    @property
    def empty(self):
        return 'No Kernels !'

    def color(self, item):
        if item == 'txt':
            return self.app.c_kern_txt
        elif item == 'bdr':
            return self.app.c_kern_bdr | curses.A_BOLD
        elif item == 'ttl':
            return self.app.c_kern_ttl | curses.A_BOLD
        elif item == 'hh':
            return self.app.c_kern_hh | curses.A_BOLD
        elif item == 'co':
            return self.app.c_kern_co | curses.A_BOLD
        elif item == 'al':
            return self.app.c_kern_al | curses.A_BOLD
        elif item == 'di':
            return self.app.c_kern_di | curses.A_BOLD
        elif item == 'pwf':
            return self.app.c_kern_pwf | curses.A_BOLD

    def get_items(self):
        """ Get items ! """

        self.app.cf = self.app.kc.connection_file

        return kernel_dic(self.app.cf)

    def custom_key_bindings(self):
        """ Key actions """

        # Menu EXPLORER
        if self.pkey in [9, 69, ord('\t')]:    # -> TAB/E
            self.app.kernel_switch = True

    def update_connection(self):
        """ Return cf and kc """

        return self.app.cf, self.app.kc

    def create_menu(self):
        """ Create the item list for the kernel menu  """

        if self.item_dic[self.selected]['type'] == 'Connected':
            return [('New', 'self._new_k()'),
                    ('Remove all died', 'self._rm_all_cf()'),
                    ('Shutdown all alive', 'self._kill_all_k()')]

        elif self.item_dic[self.selected]['type'] == 'Alive':
            return [('Connect', 'self._connect_k()'),
                    ('New', 'self._new_k()'),
                    ('Shutdown', 'self._kill_k()'),
                    ('Shutdown all alive', 'self._kill_all_k()'),
                    ('Remove all died', 'self._rm_all_cf()')]

        elif self.item_dic[self.selected]['type'] == 'Died':
            return [('Restart', 'self._restart_k()'),
                    ('New', 'self._new_k()'),
                    ('Remove file', 'self._rm_cf()'),
                    ('Remove all died', 'self._rm_all_cf()'),
                    ('Shutdown all alive', 'self._kill_all_k()')]

        else:
            return []

    def _new_k(self):
        """ Create a new kernel. """

        kid = start_new_kernel(version=self.app.config['kernel version']['version'])
        self.app.wng.display('Kernel id {} created (Python {})'.format(kid,
                                                                       self.app.config['kernel version']['version']))

    def _connect_k(self):
        """ Connect to a kernel. """

        km, self.app.kc = connect_kernel(self.item_dic[self.selected]['value'])
        send_msg(self.sock.RequestSock, '<cf>' + self.item_dic[self.selected]['value'])

        # Update kernels connection file and set new kernel flag
        self.app.cf = self.app.kc.connection_file
        self.app.kernel_change = True
        self.app.kernel_switch = True

    def _restart_k(self):
        """ Restart a died kernel. """

        self.app.wng.display("Not implemented yet")
        #  restart_kernel(self.item_dic[self.selected]['value'])

    def _kill_k(self):
        """ Kill kernel. """

        shutdown_kernel(self.item_dic[self.selected]['value'])
        self.position = 1
        self.page = 1

    def _kill_all_k(self):
        """ Kill all kernel marked as Alive. """

        for name in self.item_dic:
            if self.item_dic[name]['type'] == 'Alive':
                shutdown_kernel(self.item_dic[name]['value'])
        self.page = 1
        self.position = 1  # Reinit cursor location

    def _rm_cf(self):
        """ Remove connection file of died kernel. """

        os.remove(self.item_dic[self.selected]['value'])
        self.page = 1
        self.position = 1  # Reinit cursor location

    def _rm_all_cf(self):
        """ Remove connection files of all died kernels. """

        for name in self.item_dic:
            if self.item_dic[name]['type'] == 'Died':
                os.remove(self.item_dic[name]['value'])
        self.page = 1
        self.position = 1  # Reinit cursor location



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Mon Nov 14 09:08:25 2016
# Last Modified : jeu. 29 mars 2018 23:38:43 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
from cpyvke.objects.panel import ListPanel


class ClassWin(ListPanel):
    """ Class inspector panel """

    def __init__(self, app, sock, logger, varval, varname):
        super(ClassWin, self).__init__(app, sock, logger)
        self.varval = varval
        self.varname = varname

    @property
    def title(self):
        return ' {} inspection '.format(self.varname)

    @property
    def panel_name(self):
        return 'class-inspector'

    @property
    def empty(self):
        return 'No objects in this class instance'

    def color(self, item):

        if item == 'txt':
            return self.app.c_exp_txt
        elif item == 'bdr':
            return self.app.c_exp_bdr | curses.A_BOLD
        elif item == 'ttl':
            return self.app.c_exp_ttl | curses.A_BOLD
        elif item == 'hh':
            return self.app.c_exp_hh | curses.A_BOLD
        elif item == 'pwf':
            return self.app.c_exp_pwf | curses.A_BOLD

    def get_items(self):
        """ List of items :

            varval : generated by utils.inspector.inspect_class_instance()

        """

        return self.varval



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mar. 13 mars 2018 14:48:27 CET
# Last Modified : lun. 09 avril 2018 22:51:01 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


import curses
import logging

logger = logging.getLogger("cpyvke")


class Colors:

    def __init__(self, config):
        """ Colors class constructor """

        self.config = config
        curses.start_color()        #
        curses.setupterm()

        self.warning_colors()
        self.main_and_bar_colors()
        self.explorer_colors()
        self.kernel_colors()

    def check_color(self, color):
        """ Check if a color is set to transparent. """

        if color == 'transparent':
            curses.use_default_colors()
            return -1
        elif color.isdigit():
            if curses.COLORS > 8:
                return int(color)
            else:
                logger.error('TERM accept only 8 colors')
                return eval('curses.COLOR_UNDEFINED')
        else:
            return eval('curses.COLOR_' + color.upper())

    def warning_colors(self):
        """ Define warning colors """

        wgtxt = self.config['wg']['txt'].replace(' ', '').split(',')
        wgbdr = self.config['wg']['bdr'].replace(' ', '').split(',')
        try:
            wgtxt_fg = self.check_color(wgtxt[0])
            wgtxt_bg = self.check_color(wgtxt[1])
            wgbdr_fg = self.check_color(wgbdr[0])
            wgbdr_bg = self.check_color(wgbdr[1])
        except Exception as err:
            wgtxt_fg = curses.COLOR_RED
            wgtxt_bg = -1
            wgbdr_fg = curses.COLOR_RED
            wgbdr_bg = -1
            logger.error('%s', err)

        # Define pairs :
        curses.init_pair(1, wgtxt_fg, wgtxt_bg)
        curses.init_pair(2, wgbdr_fg, wgbdr_bg)

    def main_and_bar_colors(self):
        """ Define Main & Bar colors """

        mntxt = self.config['mn']['txt'].replace(' ', '').split(',')
        mnbdr = self.config['mn']['bdr'].replace(' ', '').split(',')
        mnttl = self.config['mn']['ttl'].replace(' ', '').split(',')
        mnhh = self.config['mn']['hh'].replace(' ', '').split(',')
        mnasc = self.config['mn']['asc'].replace(' ', '').split(',')
        try:
            mntxt_fg = self.check_color(mntxt[0])
            mntxt_bg = self.check_color(mntxt[1])
            mnbdr_fg = self.check_color(mnbdr[0])
            mnbdr_bg = self.check_color(mnbdr[1])
            mnttl_fg = self.check_color(mnttl[0])
            mnttl_bg = self.check_color(mnttl[1])
            mnhh_fg = self.check_color(mnhh[0])
            mnhh_bg = self.check_color(mnhh[1])
            mnasc_fg = self.check_color(mnasc[0])
            mnasc_bg = self.check_color(mnasc[1])

        except Exception as err:
            mntxt_fg = curses.COLOR_WHITE
            mntxt_bg = -1
            mnbdr_fg = curses.COLOR_WHITE
            mnbdr_bg = -1
            mnttl_fg = curses.COLOR_WHITE
            mnttl_bg = -1
            mnhh_fg = curses.COLOR_BLACK
            mnhh_bg = curses.COLOR_WHITE
            mnasc_fg = curses.COLOR_RED
            mnasc_bg = -1
            logger.error('%s', err)

        brkn = self.config['br']['kn'].replace(' ', '').split(',')
        brhlp = self.config['br']['hlp'].replace(' ', '').split(',')
        brco = self.config['br']['co'].replace(' ', '').split(',')
        brdco = self.config['br']['dco'].replace(' ', '').split(',')
        try:
            brkn_fg = self.check_color(brkn[0])
            brkn_bg = self.check_color(brkn[1])
            brhlp_fg = self.check_color(brhlp[0])
            brhlp_bg = self.check_color(brhlp[1])
            brco_fg = self.check_color(brco[0])
            brco_bg = self.check_color(brco[1])
            brdco_fg = self.check_color(brdco[0])
            brdco_bg = self.check_color(brdco[1])
        except Exception as err:
            brkn_fg = curses.COLOR_WHITE
            brkn_bg = -1
            brhlp_fg = curses.COLOR_WHITE
            brhlp_bg = -1
            brco_fg = curses.COLOR_GREEN
            brco_bg = -1
            brdco_fg = curses.COLOR_RED
            brdco_bg = -1
            logger.error('%s', err)

        # Define Pairs
        curses.init_pair(11, mntxt_fg, mntxt_bg)
        curses.init_pair(12, mnbdr_fg, mnbdr_bg)
        curses.init_pair(13, mnttl_fg, mnttl_bg)
        curses.init_pair(14, mnhh_fg, mnhh_bg)
        curses.init_pair(16, mnasc_fg, mnasc_bg)
        if mnttl_bg != -1:
            curses.init_pair(15, mnttl_bg, mnbdr_bg)
        else:
            curses.init_pair(15, mnbdr_fg, mnbdr_bg)

        curses.init_pair(41, brkn_fg, brkn_bg)
        curses.init_pair(42, brhlp_fg, brhlp_bg)
        curses.init_pair(43, brco_fg, brco_bg)
        curses.init_pair(44, brdco_fg, brdco_bg)
        curses.init_pair(47, mnbdr_fg, brkn_bg)

        if brdco_bg != -1:
            curses.init_pair(40, brdco_bg, mnbdr_bg)
        else:
            curses.init_pair(40, mnbdr_fg, mnbdr_bg)

        if brco_bg != -1:
            curses.init_pair(49, brco_bg, brkn_bg)
        else:
            curses.init_pair(49, mnbdr_fg, brkn_bg)

        if brkn_bg != -1:
            curses.init_pair(45, brkn_bg, mnbdr_bg)
            curses.init_pair(48, brkn_bg, brco_bg)
        else:
            curses.init_pair(45, mnbdr_fg, mnbdr_bg)
            curses.init_pair(48, mnbdr_fg, brco_bg)

        if brhlp_bg != -1:
            curses.init_pair(46, brhlp_bg, mnbdr_bg)
        else:
            curses.init_pair(46, mnbdr_fg, mnbdr_bg)

    def explorer_colors(self):
        """ Define explorer window colors """

        xptxt = self.config['xp']['txt'].replace(' ', '').split(',')
        xpbdr = self.config['xp']['bdr'].replace(' ', '').split(',')
        xpttl = self.config['xp']['ttl'].replace(' ', '').split(',')
        xphh = self.config['xp']['hh'].replace(' ', '').split(',')
        try:
            xptxt_fg = self.check_color(xptxt[0])
            xptxt_bg = self.check_color(xptxt[1])
            xpbdr_fg = self.check_color(xpbdr[0])
            xpbdr_bg = self.check_color(xpbdr[1])
            xpttl_fg = self.check_color(xpttl[0])
            xpttl_bg = self.check_color(xpttl[1])
            xphh_fg = self.check_color(xphh[0])
            xphh_bg = self.check_color(xphh[1])
        except Exception as err:
            xptxt_fg = curses.COLOR_WHITE
            xptxt_bg = -1
            xpbdr_fg = curses.COLOR_CYAN
            xpbdr_bg = -1
            xpttl_fg = curses.COLOR_CYAN
            xpttl_bg = -1
            xphh_fg = curses.COLOR_BLACK
            xphh_bg = curses.COLOR_CYAN
            logger.error('%s', err)

        # Define pairs
        curses.init_pair(21, xptxt_fg, xptxt_bg)
        curses.init_pair(22, xpbdr_fg, xpbdr_bg)
        curses.init_pair(23, xpttl_fg, xpttl_bg)
        curses.init_pair(24, xphh_fg, xphh_bg)
        if xpttl_bg != -1:
            curses.init_pair(25, xpttl_bg, xpbdr_bg)
        else:
            curses.init_pair(25, xpbdr_fg, xpbdr_bg)

    def kernel_colors(self):
        """ Define kernel window colors """
        kntxt = self.config['kn']['txt'].replace(' ', '').split(',')
        knbdr = self.config['kn']['bdr'].replace(' ', '').split(',')
        knttl = self.config['kn']['ttl'].replace(' ', '').split(',')
        knhh = self.config['kn']['hh'].replace(' ', '').split(',')
        knco = self.config['kn']['co'].replace(' ', '').split(',')
        kndi = self.config['kn']['di'].replace(' ', '').split(',')
        knal = self.config['kn']['al'].replace(' ', '').split(',')
        try:
            kntxt_fg = self.check_color(kntxt[0])
            kntxt_bg = self.check_color(kntxt[1])
            knbdr_fg = self.check_color(knbdr[0])
            knbdr_bg = self.check_color(knbdr[1])
            knttl_fg = self.check_color(knttl[0])
            knttl_bg = self.check_color(knttl[1])
            knhh_fg = self.check_color(knhh[0])
            knhh_bg = self.check_color(knhh[1])
            knco_fg = self.check_color(knco[0])
            knco_bg = self.check_color(knco[1])
            kndi_fg = self.check_color(kndi[0])
            kndi_bg = self.check_color(kndi[1])
            knal_fg = self.check_color(knal[0])
            knal_bg = self.check_color(knal[1])
        except Exception as err:
            kntxt_fg = curses.COLOR_RED
            kntxt_bg = -1
            knbdr_fg = curses.COLOR_RED
            knbdr_bg = -1
            knttl_fg = curses.COLOR_RED
            knttl_bg = -1
            knhh_fg = curses.COLOR_WHITE
            knhh_bg = -1
            knco_fg = curses.COLOR_GREEN
            knco_bg = -1
            knal_fg = curses.COLOR_CYAN
            knal_bg = -1
            kndi_fg = curses.COLOR_RED
            kndi_bg = -1
            logger.error('%s', err)

        # Define pairs
        curses.init_pair(31, kntxt_fg, kntxt_bg)
        curses.init_pair(32, knbdr_fg, knbdr_bg)
        curses.init_pair(33, knttl_fg, knttl_bg)
        curses.init_pair(34, knhh_fg, knhh_bg)
        curses.init_pair(35, knco_fg, knco_bg)
        curses.init_pair(36, knal_fg, knal_bg)
        curses.init_pair(37, kndi_fg, kndi_bg)
        if knttl_bg != -1:
            curses.init_pair(38, knttl_bg, knbdr_bg)
        else:
            curses.init_pair(38, knbdr_fg, knbdr_bg)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mar. 10 avril 2018 17:07:53 CEST
# Last Modified : mar. 10 avril 2018 20:49:40 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


__version__ = "1.2.5"



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : jeu. 01 mars 2018 15:09:06 CET
# Last Modified : mar. 27 mars 2018 23:22:52 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


import struct


def send_msg(sock, msg):
    """ Prefix each message with a 4-byte length (network byte order) """
    msg = struct.pack('>I', len(msg)) + msg.encode('utf8')
    sock.sendall(msg)


def recv_msg(sock):
    """ Read message length and unpack it into an integer """
    raw_msglen = recv_all(sock, 4)
    if not raw_msglen:
        return None
    msglen = struct.unpack('>I', raw_msglen)[0]
    # Read the message data
    return recv_all(sock, msglen)


def recv_all(sock, n):
    """ Helper function to recv n bytes or return None if EOF is hit """
    data = b''
    while len(data) < n:
        packet = sock.recv(n - len(data))
        if not packet:
            return None
        data += packet
    return data



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mer. 28 mars 2018 15:35:51 CEST
# Last Modified : mer. 28 mars 2018 15:36:50 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

RED = "\033[1;31m"
BLUE = "\033[1;34m"
CYAN = "\033[1;36m"
GREEN = "\033[0;32m"
RESET = "\033[0;0m"
BOLD = "\033[;1m"
REVERSE = "\033[;7m"



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Fri Nov 4 21:49:15 2016
# Last Modified : mer. 28 mars 2018 23:08:44 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


from jupyter_client import BlockingKernelClient, manager
import os
import subprocess
import psutil
import logging
import socket
import time

logger = logging.getLogger("cpyvke.ktools")


def start_new_kernel(LogDir=os.path.expanduser("~") + "/.cpyvke/", version=3):
    """ Start a new kernel and return the kernel id """

    with open(LogDir + 'kd5.lock', "w") as f:
        if version == '2':
            subprocess.Popen(["ipython", "kernel"], stdout=f)
        else:
            subprocess.Popen(["ipython3", "kernel"], stdout=f)

    time.sleep(1)

    with open(LogDir + 'kd5.lock', "r") as f:
        kid = f.read().split('kernel-')[1].split('.json')[0]

    with open(LogDir + 'kd5.lock', "w") as f:
        f.write(kid)

    logger.info('Create :: Kernel id. {}'.format(kid))

    return kid


def is_runing(cf):
    """ Check if kernel is alive.
    """

    kc = BlockingKernelClient()
    kc.load_connection_file(cf)
    port = kc.get_connection_info()['iopub_port']

    # if check_server(port):
    if is_open("127.0.0.1", port):
        return True
    else:
        return False


def check_server(port):
    """ Check if a service is listening on port.

    NOTE : Too slow for curses interface -> replaced by socket tests : is_open()

    """

    addr = [item.laddr for item in psutil.net_connections('inet') if str(port) in str(item.laddr[1])]
    if addr:
        return True
    else:
        return False


def is_open(ip, port):
    """ Check if port is open on ip """

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        s.shutdown(2)
        return True
    except socket.error:
        return False


def set_kid(cf):
    return cf.split('kernel-')[1].split('.json')[0]


def kernel_list(cf=None):
    """ List of connection files. """

    path = os.path.expanduser('~/.local/share/jupyter/runtime/')
    lstk1 = [path + item for item in os.listdir(path) if 'kernel' in item]
    path = '/run/user/1000/jupyter/'
    lstk2 = [path + item for item in os.listdir(path) if 'kernel' in item]
    lstk = lstk1 + lstk2

    try:
        lst = [(item, '[Alive]' if is_runing(item) else '[Died]') for item in lstk]
    except Exception:
        logger.error('No kernel available', exc_info=True)
        return []
    else:
        return [(cf, '[Connected]') if cf in item else item for item in lst]


def kernel_dic(cf=None):
    """ Dictionnary of connection files. The keys are :
        {'name': {'value': val, 'type': 'type'}}
    """

    path = os.path.expanduser('~/.local/share/jupyter/runtime/')
    lstk1 = [path + item for item in os.listdir(path) if 'kernel' in item]
    path = '/run/user/1000/jupyter/'
    lstk2 = [path + item for item in os.listdir(path) if 'kernel' in item]
    lstk = lstk1 + lstk2

    try:
        return {set_kid(item): {'value': item, 'type': 'Connected'} if is_runing(item) and item == cf
                else {'value': item, 'type': 'Alive'} if is_runing(item)
                else {'value': item, 'type': 'Died'} for item in lstk}
    except Exception:
        logger.error('No kernel available', exc_info=True)
        return {}


def print_kernel_list():
    """ Display kernel list. """
    klst = kernel_list()
    print('{:-^79}'.format('| List of available kernels |'))

    if not klst:
        print('{:^79}'.format('No kernel available'))
        print('{:^79}'.format('Last run of the daemon may have quit prematurely.'))
    else:
        for item in klst:
            print('{0[0]:71}{0[1]:>8}'.format(item))
    print(79*'-')


def print_kernel_dic():
    """ Display kernel list. """
    klst = kernel_dic()
    print('{:-^79}'.format('| List of available kernels |'))

    if not klst:
        print('{:^79}'.format('No kernel available'))
        print('{:^79}'.format('Last run of the daemon may have quit prematurely.'))
    else:
        for item in klst:
            print("{0[value]:71}{0[type]:>8}".format(klst[item]))
    print(79*'-')


def connect_kernel(cf):
    """ Connect a kernel. """

    if is_runing(cf):
        kc = BlockingKernelClient(connection_file=cf)
        kc.load_connection_file(cf)
        km = None

    else:
        # Kernel manager
        km = manager.KernelManager(connection_file=cf)
        km.start_kernel()
        # Kernel Client
        kc = km.blocking_client()

    init_kernel(kc)

    return km, kc


def connect_kernel_as_manager(cf):
    """ Connect a kernel """

    # Kernel manager
    km = manager.KernelManager(connection_file=cf)
    km.start_kernel()
    # Kernel Client
    kc = km.blocking_client()
    init_kernel(kc)

    return km, kc


def init_kernel(kc, backend='tk'):
    """ init communication. """

    backend = 'tk'

    kc.execute("import numpy as _np", store_history=False)
    kc.execute("_np.set_printoptions(threshold='nan')", store_history=False)
    kc.execute("%matplotlib {}".format(backend), store_history=False)
    kc.execute("import cpyvke.utils.inspector as _inspect", store_history=False)


def shutdown_kernel(cf):
    """ Shutdown a kernel based on its connection file. """

    km, kc = connect_kernel_as_manager(cf)
    km.shutdown_kernel(now=True)


def restart_kernel(cf):
    """ Restart a kernel based on its connection file. """

    km, kc = connect_kernel_as_manager(cf)
    try:
        km.start_kernel()
    except Exception:
        logger.error('Issue restarting kernel', exc_info=True)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov 9 10:03:04 2016
# Last Modified : ven. 30 mars 2018 17:28:02 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
import locale
from curses import panel

from cpyvke.curseswin.classwin import ClassWin
from cpyvke.curseswin.widgets import Viewer
from cpyvke.utils.display import whos_to_dic
from cpyvke.utils.comm import recv_msg
from cpyvke.utils.inspector import ProceedInspection, Inspect
from cpyvke.objects.panel import ListPanel


locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


class ExplorerWin(ListPanel):
    """ Variable explorer panel """

    def __init__(self, app, sock, logger):
        super(ExplorerWin, self).__init__(app, sock, logger)

    @property
    def panel_name(self):
        return 'variable-explorer'

    @property
    def title(self):
        return 'Variable Explorer'

    @property
    def empty(self):
        return 'Interactive namespace is empty'

    def color(self, item):
        if item == 'txt':
            return self.app.c_exp_txt
        elif item == 'bdr':
            return self.app.c_exp_bdr | curses.A_BOLD
        elif item == 'ttl':
            return self.app.c_exp_ttl | curses.A_BOLD
        elif item == 'hh':
            return self.app.c_exp_hh | curses.A_BOLD
        elif item == 'pwf':
            return self.app.c_exp_pwf | curses.A_BOLD

    def custom_tasks(self):
        """ Additional Explorer tasks """

        # Update variable number in bottom bar:
        self.app.var_nb = len(self.item_dic)

    def custom_key_bindings(self):
        """ Key Actions ! """

        # Menu KERNEL
        if self.pkey in [9, 75, ord('\t')]:    # -> TAB/K
            self.app.explorer_switch = True

        # Force Update Variable List sending fake code to daemon
        elif self.pkey == 114:   # -> r
            self.sock.force_update(self.app.wng)

    def get_items(self):
        """ Get variable from the daemon """

        try:
            tmp = recv_msg(self.sock.MainSock).decode('utf8')
        except BlockingIOError:     # If no message !
            pass
        except OSError:             # If user disconnect cpyvke from socket
            pass
        except AttributeError:      # If kd5 is stopped
            pass
        else:
            if tmp:
                self.item_dic = whos_to_dic(tmp)
                self.logger.info('Variable list updated')
                self.logger.debug('\n%s', tmp)
                try:
                    # remove temporary file used by daemon from the list
                    del self.item_dic['fcpyvke0']
                except KeyError:
                    pass

        return self.item_dic

    def menu_special_init(self):
        """ Additionnal menu init """

        # Save directory
        self.save_dir = self.app.config['path']['save-dir']

        # Variables properties
        self.varname = self.item_keys[self.position]
        self.vartype = self.item_dic[self.item_keys[self.position]]['type']
        self.varval = self.item_dic[self.item_keys[self.position]]['value']

        # Get Variable characteristics
        proc = ProceedInspection(self.app, self.sock, self.logger,
                                 self.varname, self.varval, self.vartype,
                                 self.position, self.page)
        self._ismenu, self.varname, self.varval, self.vartype, self.doc = proc.get_variable()

        # Init all Inspectors
        self.inspect = Inspect(self.varval, self.varname, self.vartype)
        self.class_win = ClassWin(self.app, self.sock, self.logger, self.varval, self.varname)
        self.view = Viewer(self.app, self.varval, self.varname)
        if self.doc:
            self.view_doc = Viewer(self.app, self.doc, self.varname)
            self.inspect_doc = Inspect(self.doc, self.varname, self.vartype)

    def create_menu(self):
        """ Create the item list for the general menu. """

        if self.varval == '[Busy]':
            return [('Busy', ' ')]

        elif self.vartype in self.inspect.type_numeric():
            return [('Delete', "self.sock.del_var(self.varname, self.app.wng)")]

        elif self.vartype in ['module']:
            return [('Help', "self.inspect.display('less', 'help')")]

        elif self.vartype in ['function']:
            return [('Help', 'self.view_doc.display()'),
                    ('Code', 'self.view.display()'),
                    ('Less help', "self.inspect_doc.display('less', 'help')"),
                    ('Less code', "self.inspect.display('less', 'help')")]

        elif self.vartype in ['builtin_function_or_method']:
            return [('Help', 'self.view_doc.display()'),
                    ('Less', "self.inspect_doc.display('less', 'help')")]

        elif self.vartype in self.inspect.type_struct():
            return [('View', 'self.view.display()'),
                    ('Less', "self.inspect.display('less')"),
                    ('Edit', "self.inspect.display('vim')"),
                    ('Save', "self.item_save()"),
                    ('Delete', "self.sock.del_var(self.varname, self.app.wng)")]

        elif (self.vartype == 'ndarray') and (len(self.varval.shape) == 1):
            return [('Plot', 'self.inspect.plot1D()'),
                    ('Save', 'self.menu_save()'),
                    ('Delete', "self.sock.del_var(self.varname, self.app.wng)")]

        elif (self.vartype == 'ndarray') and (len(self.varval.shape) == 2):
            return [('Plot 2D', 'self.inspect.plot2D()'),
                    ('Plot (cols)', 'self.inspect.plot1Dcols()'),
                    ('Plot (lines)', 'self.inspect.plot1Dlines()'),
                    ('Save', 'self.menu_save()'),
                    ('Delete', "self.sock.del_var(self.varname, self.app.wng)")]

        elif self.vartype in ['class', 'type']:
            return [('View', 'self.view.display()'),
                    ('Inspect', 'self.class_win.display()')]

        else:
            return [('No Option', 'exit')]

    def item_save(self):
        """ Item save in menu """

        self.inspect.save(self.save_dir)
        self.app.wng.display('Saved !')

    def menu_save(self):
        """ Create the save menu. """

        # Init Menu
        save_menu = self.app.stdscr.subwin(5, 6, self.menu_height, self.app.screen_width-9)
        save_menu.bkgd(self.color('txt'))
        save_menu.attrset(self.color('bdr'))
        save_menu.border(0)
        save_menu.keypad(1)

        # Send menu to a panel
        panel_save = panel.new_panel(save_menu)
        panel_save.hide()

        # Various variables
        self.menu_cursor = 0

        save_lst = [('txt', "self.inspect.save_np(self.varname, self.save_dir, 'txt')"),
                    ('npy', "self.inspect.save_np(self.varname, self.save_dir, 'npy')"),
                    ('npz', "self.inspect.save_np(self.varname, self.save_dir, 'npz')")]
        panel_save.top()        # Push the panel to the bottom of the stack.
        panel_save.show()       # Display the panel (which might have been hidden)
        save_menu.clear()

        menukey = -1
        while menukey not in self.app.kquit:

            self.logger.error('Save items : {}'.format(len(save_lst)))
            self.logger.error('Menu Cursor : {}'.format(self.menu_cursor))
            self.logger.error('Menu Key : {}'.format(menukey))

            save_menu.border(0)
            save_menu.refresh()

            # Fill the menu
            for index, item in enumerate(save_lst):
                if index == self.menu_cursor:
                    mode = self.color('hh')
                else:
                    mode = self.color('txt') | curses.A_DIM

                save_menu.addstr(1+index, 1, item[0], mode)

            menukey = save_menu.getch()

            if menukey in self.app.kenter:
                try:
                    eval(save_lst[self.menu_cursor][1])
                    self.app.wng.display('Saved !')
                except Exception:
                    self.app.wng.display('Not saved !')
                    self.logger.error('Menu save', exc_info=True)
                else:
                    self.menu_cursor = 0
                    break

            elif menukey in self.app.kup:
                self.navigate_menu(-1, len(save_lst))

            elif menukey in self.app.kdown:
                self.navigate_menu(1, len(save_lst))

            elif menukey == curses.KEY_RESIZE:
                self.menu_cursor = 0
                break

        save_menu.clear()
        panel_save.hide()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Mon Nov 14 09:08:25 2016
# Last Modified : mar. 03 avril 2018 15:55:45 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import time
import locale
import curses
import abc
from curses import panel
from math import ceil
from cpyvke.curseswin.widgets import Help
from cpyvke.curseswin.prompt import Prompt
from cpyvke.curseswin.app import check_size
from cpyvke.utils.kd import restart_daemon
from cpyvke.utils.display import format_cell
from cpyvke.utils.comm import send_msg

code = locale.getpreferredencoding()


class BasePanel(abc.ABC):
    """ Generic Panel.
    """

    def __init__(self, app, sock, logger):
        """ Class Constructor """

        # Instance arguments
        self.app = app
        self.sock = sock
        self.logger = logger

        # Init constants
        self.resize = False
        self.pkey = -1

        # Init Prompt
        self.prompt = Prompt(self.app)
        self.prompt_time = 0
        self.prompt_msg = ''

        # Update dimensions
        self.screen_height, self.screen_width = self.app.stdscr.getmaxyx()  # Local dimensions

        # Init subwin
        self.gwin = self.app.stdscr.subwin(self.app.panel_height, self.app.screen_width, 0, 0)
        self.gwin.keypad(1)

        # Init Panel
        self.gpan = panel.new_panel(self.gwin)
        self.gpan.hide()

    @property
    @abc.abstractmethod
    def title(self):
        """ Panel title. Must return a string """

    @property
    @abc.abstractmethod
    def panel_name(self):
        """ Panel reference name. Must return a string """

    @abc.abstractmethod
    def color(self, item):
        """ Panel colors. Required :
            * for BasePanel : 'txt', 'bdr', 'ttl', 'hh', 'pwf'
            * for ListPanel : 'co', 'al', 'di'
        """

    @abc.abstractmethod
    def fill_main_box(self):
        """ Fill the main box """

    def display(self):
        """ Display the panel. """

        try:
            self.pkey = -1
            while self.app.close_signal == 'continue':
                self.tasks()
            self.app.shutdown()
        except Exception:
            self.app.exit_with_error()

    @check_size
    def tasks(self):
        """ List of tasks at each iteration """

        self.resize_curses()

        # Check switch panel
        if self.app.explorer_switch:
            self.app.explorer_switch = False
            self.app.kernel_win.display()
            self.resize_curses(True)

        elif self.app.kernel_switch:
            self.app.kernel_switch = False
            self.app.explorer_win.display()
            self.resize_curses(True)

        else:
            # Check Connection to daemon
            self.sock.check_main_socket()

            # Keys
            self.common_key_bindings()

            # Decrease delay right here to avoid some waiting at the getch when not
            # in switch mode. If switch, the halfdelay is set to its normal value
            # just after, in the refresh() method !
            curses.halfdelay(1)

            # Skip end of tasks if switching panel !
            if not self.app.explorer_switch and not self.app.kernel_switch and self.app.close_signal == "continue":

                # Update screen size if another menu break because of resizing
                self.resize_curses()

                # Update all static panels
                self.refresh()

            # Get pressed key (even in case of switch)
            self.pkey = self.app.stdscr.getch()

    def refresh(self):
        """ Refresh all objects. """

        # Erase all windows
        self.gwin.erase()
        self.app.stdscr.erase()

        # Create border before updating fields
        self.gwin.border(0)

        # Fill the main box !
        self.fill_main_box()

        # Update all windows
        if self.app.debug:
            self.app.dbg_pad(self.pkey)

        # Update infos -- Bottom
        self.app.status_bar()
        self.prompt_msg_display()
        self.app.stdscr.refresh()
        self.gwin.refresh()

        # Reactive timeout for getch
        curses.halfdelay(self.app.curse_delay)

    def common_key_bindings(self):
        """ Common key bindings """

        # Custom key bindings
        self.custom_key_bindings()

        # Socket actions
        self.socket_key_bindings()

        # Item list
        self.list_key_bindings()

        # Menu Help
        if self.pkey == 63:    # -> ?
            help_menu = Help(self.app)
            help_menu.display()

        # Prompt
        elif self.pkey == 58:     # -> :
            self.cmd = self.prompt.with_completion(chr(self.pkey))
            self.prompt_cmd()

        # Send code
        elif self.pkey == 120:    # -> x
            self.send_code()

        # Debug Pad
        elif self.pkey == 100:      # -> d
            self.toggle_debug()

    def socket_key_bindings(self):
        """ Socket actions key bindings. """

        if self.pkey == 82:    # -> R
            self.daemon_restart_connection()

        elif self.pkey == 68:    # -> D
            self.daemon_disconnect()

        elif self.pkey == 67:     # -> C
            self.daemon_connect()

        elif self.pkey == 18:    # -> c-r
            self.daemon_restart()

    def list_key_bindings(self):
        """ Not available for BasePanel. See List ListPanel """

        pass

    def custom_key_bindings(self):
        """ Key bindings : To overload """

        pass

    def prompt_msg_display(self):
        """ Erase prompt message after some delay """

        if self.prompt_msg and time.time() - self.prompt_time > 3:
            self.prompt_msg = ''
        else:
            self.prompt.display(self.prompt_msg)

    def prompt_msg_setup(self, msg):
        """ Set up the message to display in the prompt """

        self.prompt_msg = msg
        self.prompt_time = time.time()

    def prompt_cmd(self):
        """ Actions for prompt """

        if not self.cmd:
            pass

        elif self.cmd in ["q", "quit"]:
            self.app.close_signal = 'close'

        elif self.cmd in ["Q", "Quit"]:
            self.app.close_signal = 'shutdown'

        elif self.cmd in ['k', 'K', 'kernel-manager']:
            self.prompt_cmd_kernel_manager()

        elif self.cmd in ['v', 'V', 'e', 'E', 'variable-explorer']:
            self.prompt_cmd_variable_explorer()

        elif self.cmd in ['h', 'help']:
            help_menu = Help(self.app)
            help_menu.display()

        elif self.cmd in ['R', 'daemon-restart']:
            self.daemon_restart()

        elif self.cmd in ['r', 'daemon-restart-connection']:
            self.daemon_restart_connection()

        elif self.cmd in ['c', 'daemon-connect']:
            self.daemon_connect()

        elif self.cmd in ['d', 'daemon-disconnect']:
            self.daemon_disconnect()

        elif self.cmd in ['toggle-debug']:
            self.toggle_debug()

        else:
            self.prompt_msg_setup('Command not found !')

    def prompt_cmd_kernel_manager(self):
        """ 'kernel-manager' prompt command"""

        if self.panel_name in ['variable-explorer']:
            self.app.explorer_switch = True
        elif self.panel_name not in ['kernel-manager']:
            self.app.kernel_win.display()
        else:
            self.prompt_msg_setup('Already in kernel manager !')

    def prompt_cmd_variable_explorer(self):
        """ 'variable-explorer' prompt command """

        if self.panel_name in ['kernel-manager']:
            self.app.kernel_switch = True
        elif self.panel_name not in ['variable-explorer']:
            self.app.explorer_win.display()
        else:
            self.prompt_msg_setup('Already in variable explorer !')

    def toggle_debug(self):
        """ Display/hide debug informations """

        if self.app.debug:
            self.app.debug = False
        else:
            self.app.debug = True

    def send_code(self):
        """ Send code to current kernel """

        code = self.prompt.simple('Send-code ')
        code, err = self.check_code(code)
        if err:
            self.prompt_msg_setup(err)
        elif code:
            try:
                send_msg(self.sock.RequestSock, '<code>' + code)
                self.logger.info('Code sent to kernel : {}'.format(code))
                self.prompt_msg_setup('Code sent !')
            except Exception:
                self.logger.error('Code not sent !')
                self.prompt_msg_setup('Code not sent !')

    @staticmethod
    def check_code(code):
        """ Check is code is authorized """

        if 'input' in code:
            return '', 'input command is not allowed'
        elif 'reset' in code:
            return 'reset -f', 'Resetting namespace...'
        else:
            return code, None

    def daemon_connect(self):
        """ Connect to daemon socket """

        self.sock.init_sockets()
        self.sock.warning_socket(self.app.wng)

    def daemon_disconnect(self):
        """ Disconnet from daemon socket """

        self.sock.close_sockets()
        self.sock.warning_socket(self.app.wng)

    def daemon_restart_connection(self):
        """ Restart connection to the daemon socket """

        self.app.wng.display(' Restarting connection ')
        self.sock.restart_sockets()
        self.sock.warning_socket(self.app.wng)

    def daemon_restart(self):
        """ Restart kd5 ! """

        restart_daemon()
        self.app.wng.display(' Restarting Daemon ')
        self.sock.init_sockets()
        self.sock.warning_socket(self.app.wng)

    def resize_curses(self, force=False):
        """ Check if terminal is resized and adapt screen """

        # Check difference between self.screen_height and self.app.screen_height
        resize = curses.is_term_resized(self.screen_height, self.screen_width)
        if resize or force:
            # save also these value locally to check if
            self.screen_height, self.screen_width = self.app.stdscr.getmaxyx()
            # Update display
            self.app.stdscr.clear()
            self.gwin.clear()
            self.gwin.resize(self.app.panel_height, self.app.screen_width)
            curses.resizeterm(self.app.screen_height, self.app.screen_width)
            self.app.stdscr.refresh()
            self.gwin.refresh()


class ListPanel(BasePanel):
    """ Generic Panel for lists with menu.
    """

    def __init__(self, app, sock, logger):
        """ Class Constructor """

        super(ListPanel, self).__init__(app, sock, logger)

        # Some variables
        self.filter = None
        self.mk_sort = 'name'
        self.search = None
        self.search_index = 0
        self.search_lst = []
        self.limit_msg = ''
        self.position = 0
        self.page = 1

        # Init variables :
        self.item_dic = {}
        self.item_keys = []

    @property
    @abc.abstractmethod
    def empty(self):
        """ Text for empty list. Must return a string """

        return

    def display(self):
        """ Display the panel. """

        # Init colors
        self.gwin.bkgd(self.color('txt'))
        self.gwin.attrset(self.color('bdr'))

        self.gpan.top()     # Push the panel to the bottom of the stack.
        self.gpan.show()    # Display the panel
        self.gwin.clear()

        # Update size if it has change when panel was hidden
        self.resize_curses(True)

        self.pkey = -1
        while self.pkey not in self.app.kquit and self.app.close_signal == 'continue':

            if self.app.kernel_switch or self.app.explorer_switch:
                break

            self.tasks()

        self.gwin.clear()
        self.gpan.hide()

    def custom_tasks(self):
        """ Supplementary tasks [To overload if needed] """

        pass

    @check_size
    def tasks(self):
        """ List of tasks at each iteration """

        # Listen to resize and adapt Curses
        self.resize_curses()

        # Custom tasks
        self.custom_tasks()

        # Check Connection to daemon
        self.sock.check_main_socket()

        # Get items
        self.item_dic = self.get_items()
        self.row_num = len(self.item_dic) - 1

        # Arange item list
        self.arange_lst()

        # Key bindings
        self.common_key_bindings()

        if not self.app.kernel_switch and not self.app.explorer_switch and self.app.close_signal == "continue":

            # Navigate in the variable list window
            self.navigate_lst()

            # Update screen size
            self.resize_curses()

            # Update all
            self.refresh()

            # Get key
            self.pkey = self.app.stdscr.getch()

    def refresh(self):
        """ Refresh all objects. """

        # Erase all windows
        self.gwin.erase()
        self.app.stdscr.erase()

        # Create border before updating fields
        self.gwin.border(0)

        # Fill the main box !
        self.fill_main_box()

        # Update all windows
        if self.app.debug:
            self.app.dbg_pad(self.pkey, self.search, self.filter, self.mk_sort)

        # Update infos -- Bottom
        self.app.status_bar()
        self.prompt_msg_display()
        self.app.stdscr.refresh()
        self.gwin.refresh()

        # Reactive timeout for getch
        curses.halfdelay(self.app.curse_delay)

    def list_key_bindings(self):
        """ Actions linked to list of item. """

        # Menu Search
        if self.pkey == 47:    # -> /
            self.search_item('Search for : ')

        # Next item (search)
        if self.pkey == 110:    # -> n
            self.search_item_next()

        # Sort variable by name/type
        elif self.pkey == 115:       # -> s
            if self.mk_sort == 'name':
                self.mk_sort = 'type'
            elif self.mk_sort == 'type':
                self.mk_sort = 'name'
            self.arange_lst()

        # Filter variables
        elif self.pkey == 102:    # -> f
            self.filter = self.prompt.simple('Limit to : ')
            if self.filter:
                self.mk_sort = 'filter'
                self.position = 0
                self.page = 1
                self.arange_lst()
            else:
                self.filter = None

        # Reinit
        elif self.pkey == 117:   # -> u
            self.mk_sort = 'name'
            self.limit_msg = ''
            self.position = 0
            self.page = 1
            self.arange_lst()

        # Panel Menu
        elif self.pkey in self.app.kenter and self.row_num != -1:
            self.init_menu()

    def custom_key_bindings(self):
        """ Key bindings : To overload """

        pass

    @abc.abstractmethod
    def get_items(self):
        """ Return a dicionnary with items : self.item_dic """

        return

    def fill_main_box(self):
        """ Update the item list """

        # Title
        if self.app.config['font']['pw-font'] == 'True':
            self.gwin.addstr(0, int((self.app.screen_width-len(self.title))/2),
                             '', self.color('pwf'))
            self.gwin.addstr(self.title, self.color('ttl'))
            self.gwin.addstr('', self.color('pwf'))
        else:
            self.gwin.addstr(0, int((self.app.screen_width-len(self.title))/2),
                             '|' + self.title + '|', self.color('ttl'))

        # Reset position if position is greater than the new list of var (reset)
        self.row_num = len(self.item_keys) - 1
        if self.position > self.row_num:
            self.position = 0
            self.page = 1

        # Items
        for i in range(self.app.row_max*(self.page-1),
                       self.app.row_max + self.app.row_max*(self.page-1)):

            if self.row_num == -1:
                self.gwin.addstr(1, 1, self.empty, self.color('hh'))

            elif i <= self.row_num:
                self.cell1, self.cell2 = format_cell(self.item_dic, self.item_keys[i], self.app.screen_width)
                if i == self.position:
                    self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), 2,
                                     self.cell1.encode(code), self.color('hh'))
                    self.fill_main_box_type_selected(i)
                else:
                    self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), 2,
                                     self.cell1.encode(code), curses.A_DIM | self.color('txt'))
                    self.fill_main_box_type(i)

        # Bottom info
        if self.app.config['font']['pw-font'] == 'True' and len(self.limit_msg) > 0:
            self.gwin.addstr(self.app.panel_height-1,
                             int((self.app.screen_width-len(self.limit_msg))/2),
                             '', self.color('pwf'))
            self.gwin.addstr(self.limit_msg, self.color('ttl'))
            self.gwin.addstr('',  self.color('pwf'))
        elif len(self.limit_msg) > 0:
            self.gwin.addstr(self.app.panel_height-1,
                             int((self.app.screen_width-len(self.limit_msg))/2),
                             '< ' + self.limit_msg + ' >', curses.A_DIM | self.color('ttl'))

        self.app.stdscr.refresh()
        self.gwin.refresh()

    def fill_main_box_type_selected(self, i):
        if "[Died]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('di'))
        elif "[Alive]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('al'))
        elif "[Connected]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('co'))
        else:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('hh'))

    def fill_main_box_type(self, i):
        if "[Died]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('di'))
        elif "[Alive]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('al'))
        elif "[Connected]" in self.cell2:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('co'))
        else:
            self.gwin.addstr(i + 1 - self.app.row_max*(self.page-1), len(self.cell1),
                             self.cell2, self.color('txt'))

    @staticmethod
    def filter_var_lst(item_dic, filt):
        """ Filter variable list (name|type). """

        filtered = []
        for key in list(item_dic):
            if filt in item_dic[key]['type'] or filt in key:
                filtered.append(key)

        return sorted(filtered)

    @staticmethod
    def type_sort(item_dic):
        """ Sort variable by type. """

        from operator import itemgetter

        types = []
        for key in list(item_dic):
            types.append([key, item_dic[key]['type']])

        types.sort(key=itemgetter(1))

        return [item[0] for item in types]

    def arange_lst(self):
        """ Organize/Arange variable list. """

        if self.mk_sort == 'name':
            self.item_keys = sorted(list(self.item_dic))

        elif self.mk_sort == 'type':
            self.item_keys = self.type_sort(self.item_dic)

        elif self.mk_sort == 'filter' and self.filter:
            self.item_keys = self.filter_var_lst(self.item_dic, self.filter)
            if not self.item_keys:
                self.prompt_msg_setup('{} not found'.format(self.filter))
                self.item_keys = sorted(list(self.item_dic))
                self.filter = None
                self.mk_sort = 'name'
            else:
                self.limit_msg = ' Filter : {} ({} obj.) '.format(self.filter, len(self.item_keys))

        else:
            self.item_keys = list(self.item_dic)

        # Update number of columns
        self.row_num = len(self.item_keys) - 1

    def search_item(self, txt_msg):
        """ Search an object in the variable list """

        self.search = self.prompt.simple(txt_msg)
        self.search_lst = [i for i, s in enumerate(self.item_keys) if self.search in s]
        self.search_index = 0
        self.logger.info('Searching for : {} in :\n{}'.format(self.search, self.item_keys))

        if self.search_lst and self.search:
            if len(self.search_lst) == 1:
                self.prompt_msg_setup("{} occurence of '{}' found".format(len(self.search_lst), self.search))
            else:
                self.prompt_msg_setup("{} occurences of '{}' found".format(len(self.search_lst), self.search))
            self.position = self.search_lst[self.search_index]
            self.page = ceil((self.position+1)/self.app.row_max)
        elif not self.search:
            pass
        else:
            self.prompt_msg_setup(self.search + ' not found !')
            self.position = 0
            self.page = 1

    def search_item_next(self):
        """ Next occurence of the searching. """

        self.search_lst = [i for i, s in enumerate(self.item_keys) if self.search in s]

        if self.search_lst and self.search_index < len(self.search_lst) - 1:
            self.search_index += 1
        else:
            self.search_index = 0

        self.position = self.search_lst[self.search_index]
        self.page = ceil((self.position+1)/self.app.row_max)

    def navigate_lst(self):
        """ Navigation though the item list"""

        self.pages = ceil((self.row_num + 1)/self.app.row_max)
        if self.pkey in self.app.kdown:
            self.navigate_down()
        if self.pkey in self.app.kup:
            self.navigate_up()
        if self.pkey in self.app.kleft and self.page > 1:
            self.navigate_left()
        if self.pkey in self.app.kright and self.page < self.pages:
            self.navigate_right()

    def navigate_right(self):
        """ Navigate Right. """

        self.page = self.page + 1
        self.position = self.app.row_max*(self.page-1)

    def navigate_left(self):
        """ Navigate Left. """

        self.page = self.page - 1
        self.position = self.app.row_max*(self.page-1)

    def navigate_up(self):
        """ Navigate Up. """

        if self.page == 1:
            if self.position > 0:
                self.position = self.position - 1
        else:
            if self.position > self.app.row_max*(self.page - 1):
                self.position = self.position - 1
            else:
                self.page = self.page - 1
                self.position = self.app.row_max - 1 + self.app.row_max*(self.page - 1)

    def navigate_down(self):
        """ Navigate Down. """

        # First page
        if self.page == 1:
            if (self.position < self.app.row_max - 1) and (self.position < self.row_num):
                self.position = self.position + 1
            else:
                if self.pages > 1:
                    self.page = self.page + 1
                    self.position = self.app.row_max*(self.page - 1)

        # Last page
        elif self.page == self.pages:
            if self.position < self.row_num:
                self.position = self.position + 1

        # Between
        else:
            if self.position < self.app.row_max - 1 + self.app.row_max*(self.page - 1):
                self.position = self.position + 1
            else:
                self.page = self.page + 1
                self.position = self.app.row_max*(self.page - 1)

    def init_menu(self):
        """ Init the menu """

        self.selected = self.item_keys[self.position]

        # Add specific initilization
        self.menu_special_init()

        # Create menu list
        self.menu_lst = self.create_menu()

        # Various variables
        self.menu_cursor = 0
        self.menu_title = ' ' + self.selected.split('/')[-1] + ' '

        # Menu dimensions
        self.menu_width = len(max(
            [self.menu_lst[i][0] for i in range(len(self.menu_lst))], key=len))
        self.menu_width = max(self.menu_width, len(self.menu_title)) + 4
        self.menu_height = len(self.menu_lst) + 2
        self.title_pos = int((self.menu_width - len(self.menu_title) - 2)/2)

        # Init Menu
        self.gwin_menu = self.app.stdscr.subwin(self.menu_height,
                                                self.menu_width, 2,
                                                self.app.screen_width-self.menu_width-2)
        self.gwin_menu.border(0)
        self.gwin_menu.bkgd(self.color('txt'))
        self.gwin_menu.attrset(self.color('bdr'))  # Change border color
        self.gwin_menu.keypad(1)

        # Send menu to a panel
        self.gpan_menu = panel.new_panel(self.gwin_menu)
        # Hide the panel. This does not delete the object, it just makes it invisible.
        self.gpan_menu.hide()
        panel.update_panels()

        # Submenu
        self.display_menu()

    def menu_special_init(self):
        """ Additionnal initialization for menu """

        pass

    def create_menu(self):
        """ Create the item list for the kernel menu : To overload """

        return [('No Option', 'None')]

    def display_menu(self):
        """ Display the menu """

        self.gpan_menu.top()        # Push the panel to the bottom of the stack.
        self.gpan_menu.show()       # Display the panel (which might have been hidden)
        self.gwin_menu.clear()

        menukey = -1
        while menukey not in self.app.kquit:
            self.gwin_menu.border(0)

            # Title
            if self.app.config['font']['pw-font'] == 'True':
                self.gwin_menu.addstr(0, self.title_pos, '', self.color('pwf'))
                self.gwin_menu.addstr(self.menu_title, self.color('ttl'))
                self.gwin_menu.addstr('', self.color('pwf'))
            else:
                self.gwin_menu.addstr(0, self.title_pos,
                                      '|' + self.menu_title + '|', self.color('ttl'))

            self.gwin_menu.refresh()

            # Create entries
            for index, item in enumerate(self.menu_lst):
                if index == self.menu_cursor:
                    mode = self.color('hh')
                else:
                    mode = self.color('txt') | curses.A_DIM

                self.gwin_menu.addstr(1+index, 1, item[0], mode)

            # Wait for keyboard event
            menukey = self.gwin_menu.getch()

            if menukey in self.app.kenter:
                eval(self.menu_lst[self.menu_cursor][1])
                break

            elif menukey in self.app.kup:
                self.navigate_menu(-1, len(self.menu_lst))

            elif menukey in self.app.kdown:
                self.navigate_menu(1, len(self.menu_lst))

            if menukey == curses.KEY_RESIZE:
                self.resize = True
                break

        self.gwin_menu.clear()
        self.gpan_menu.hide()

    def navigate_menu(self, n, size):
        """ Navigate through the menu """

        self.menu_cursor += n

        if self.menu_cursor < 0:
            self.menu_cursor = 0
        elif self.menu_cursor >= size:
            self.menu_cursor = size - 1



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov 9 16:29:28 2016
# Last Modified : mar. 03 avril 2018 21:15:01 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


import curses
from curses import panel
from time import sleep
import locale

from cpyvke.utils.display import dump, str_reduce, str_format
from cpyvke.objects.pad import PadWin

locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


class Viewer(PadWin):
    """ Display variable content in a pad. """

    def __init__(self, app, varval, varname):
        super(Viewer, self).__init__(app)
        self.varval = varval
        self.varname = varname

    def color(self, item):
        if item == 'txt':
            return self.app.c_exp_txt
        elif item == 'bdr':
            return self.app.c_exp_bdr | curses.A_BOLD
        elif item == 'ttl':
            return self.app.c_exp_ttl | curses.A_BOLD
        elif item == 'pwf':
            return self.app.c_exp_pwf | curses.A_BOLD

    @property
    def title(self):
        return ' ' + self.varname + ' '

    @property
    def content(self):
        if type(self.varval) is str:
            dumped = str_format(self.varval, self.app.screen_width-6)
        else:
            dumped = dump(self.varval)

        return dumped


class WarningMsg:
    """ Display a message. """

    def __init__(self, app):

        self.app = app

    def display(self, wng_msg):
        """ Display **wng_msg** in a panel. """

        # Init Menu
        wng_msg = str_reduce(wng_msg, self.app.screen_width - 2)
        wng_width = len(wng_msg) + 2
        menu_wng = self.app.stdscr.subwin(3, wng_width, 3, int((self.app.screen_width-wng_width)/2))
        menu_wng.bkgd(self.app.c_warn_txt | curses.A_BOLD)
        menu_wng.attrset(self.app.c_warn_bdr | curses.A_BOLD)    # change border color
        menu_wng.border(0)
        menu_wng.keypad(1)

        # Send menu to a panel
        panel_wng = panel.new_panel(menu_wng)
        panel_wng.top()        # Push the panel to the bottom of the stack.

        menu_wng.addstr(1, 1, wng_msg, self.app.c_warn_txt)
        panel_wng.show()       # Display the panel (which might have been hidden)
        menu_wng.refresh()
        sleep(1)

        # Erase the panel
        menu_wng.clear()
        panel_wng.hide()


class Help(PadWin):
    """ Display help in a pad. """

    def __init__(self, app):
        super(Help, self).__init__(app)

    def color(self, item):
        if item == 'txt':
            return self.app.c_main_txt
        elif item == 'bdr':
            return self.app.c_main_bdr | curses.A_BOLD
        elif item == 'ttl':
            return self.app.c_main_ttl | curses.A_BOLD
        elif item == 'pwf':
            return self.app.c_main_pwf | curses.A_BOLD

    @property
    def title(self):
        return ' Help '

    @property
    def content(self):
        return ['',
                '(?) This Help !',
                '(ENTER) Selected item menu',
                '(q|ESC) Previous menu',
                '(:) Prompt access',
                '(/) Search a pattern',
                '(n) Next occurence of pattern',
                '(s) Sort by name/type',
                '(f) filter',
                '(x) Execute code in current IPython kernel',
                '(u) Undo limit',
                '(r) Refresh explorer',
                '(E) Variable Explorer',
                '(K) Kernel Manager',
                '(TAB) Kernel <-> Variable',
                '(c-r) Restart daemon',
                '(R) Restart connection to daemon',
                '(D) Disconnect from daemon',
                '(C) Connect to daemon',
                '(↓|j) Next line',
                '(↑|k) Previous line',
                '(→|↡|l) Next page',
                '(←|↟|h) Previous page',
                '']


class suspend_curses:
    """Context Manager to temporarily leave curses mode"""

    def __enter__(self):
        curses.endwin()

    def __exit__(self, exc_type, exc_val, tb):
        stdscreen = curses.initscr()
        stdscreen.refresh()
        curses.doupdate()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date :
# Last Modified : mer. 11 avril 2018 00:15:49 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

from setuptools import setup, find_packages
import sys
import cpyvke

if sys.version_info < (2,7):
    raise NotImplementedError(
        "Sorry, you need at least Python 2.7 to use cpyvke.")

setup(
    name='cpyvke',
    description='A Kernel and variable explorer in Curses',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version=cpyvke.__version__,
    license='GPL',
    url='https://github.com/ipselium/cpyvke',
    author='Cyril Desjouy',
    author_email='ipselium@free.fr',
    packages=find_packages(),
    include_package_data=True,
    install_requires=["jupyter_client", "ipykernel", "psutil", "numpy", "matplotlib"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development :: Debuggers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    ],
    entry_points={
        'console_scripts': [
            'kd5 = cpyvke.kd5:main',
            'cpyvke = cpyvke.cpyvke:main',
            'cpyvke-launch-ipython = cpyvke.launch:main',
        ],
    }
)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mer. 28 mars 2018 23:07:12 CEST
# Last Modified : jeu. 29 mars 2018 00:09:09 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import sys
import psutil
import subprocess


def kd_status(pidfile):
    """ Check kd5 status """

    if os.path.exists(pidfile):
        if is_kd_running(pidfile):
            return read_pid(pidfile)
        else:
            os.remove(pidfile)
            return None
    else:
        pids = find_lost_pid()
        if len(pids) == 1:
            with open(pidfile, 'w') as f:
                f.write(pids[0])
            return pids[0]
        else:
            return None


def find_lost_pid():
    cur_pid = os.getpid()
    pids = []
    for proc in psutil.process_iter():
        pinfo = proc.as_dict(attrs=['pid', 'name'])
        if pinfo['name'] == 'kd5' and pinfo['pid'] != cur_pid:
            pids.append(pinfo['pid'])
    return pids


def read_pid(pidfile):
    """ Read the pid in pidfile """

    with open(pidfile, 'r') as f:
        pid = int(f.read())
    return pid


def kdread(cfile):
    """ read lockfile | pidfile """

    if os.path.exists(cfile):
        with open(cfile, "r") as f:
            return f.readline().replace('\n', '')
    else:
        message = 'Error :\tCannot find {} !\n\tExiting\n'
        sys.stderr.write(message.format(cfile))
        return None


def kdwrite(cfile, content):
    """ write lockfile | pidfile """

    with open(cfile, "w") as f:
        f.write(content)


def is_kd_running(pidfile):
    """ Check if process with pid (in pidfile) is actually running"""

    pid = read_pid(pidfile)
    if psutil.pid_exists(pid):
        return True
    else:
        return False


def restart_daemon():
    """ Restart kd5 """

    subprocess.Popen(["kd5", "restart"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov 9 10:03:04 2016
# Last Modified : lun. 09 avril 2018 22:43:21 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
import locale

from cpyvke.curseswin.kernelwin import KernelWin
from cpyvke.curseswin.explorerwin import ExplorerWin
from cpyvke.curseswin.widgets import WarningMsg
from cpyvke.objects.panel import BasePanel
from cpyvke.utils.ascii import ascii_cpyvke

locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


class MainWin(BasePanel):
    """ Main window. """

    def __init__(self, app, sock, logger):
        """ Main window constructor """

        super(MainWin, self).__init__(app, sock, logger)

        # Add explorer, kernel panels and wng pad to self.app !
        self.app.explorer_win = ExplorerWin(self.app, self.sock, self.logger)
        self.app.kernel_win = KernelWin(self.app, self.sock, self.logger)
        self.app.wng = WarningMsg(self.app)

    @property
    def title(self):
        return ''

    @property
    def panel_name(self):
        return 'main'

    def color(self, item):
        if item == 'txt':
            return self.app.c_main_txt
        elif item == 'bdr':
            return self.app.c_main_bdr
        elif item == 'ttl':
            return self.app.c_main_ttl
        elif item == 'hh':
            return self.app.c_main_hh
        elif item == 'pwf':
            return self.app.c_main_pwf
        elif item == 'asc':
            return self.app.c_main_asc

    def fill_main_box(self):
        """ Welcome message """

        msg = ascii_cpyvke(self.app.config['font']['ascii-font'])
        msg_size = max([len(i) for i in msg])

        for i in range(len(msg)):
            self.gwin.addstr(i+1,
                             int((self.app.screen_width - msg_size)/2),
                             msg[i], self.color('asc') | curses.A_BOLD)

        tmp = "Type   {:25} {:30}"
        self.gwin.addstr(i+3, 1, tmp.format(':help<Enter> or ?',
                                            'for help').center(self.app.screen_width-4))
        self.gwin.addstr(i+4, 1, tmp.format(':q<Enter>',
                                            'to exit').center(self.app.screen_width-4))
        self.gwin.addstr(i+5, 1, tmp.format(':Q<Enter>',
                                            'to exit and shutdown daemon').center(self.app.screen_width-4))

    def custom_key_bindings(self):
        """ Custom Key Actions ! """

        # Kernel Panel
        if self.pkey == 75:    # -> K
            self.app.kernel_win.display()
            self.resize_curses(True)  # Fix brutal resize crash
            if self.app.kernel_change:
                self.app.cf, self.app.kc = self.app.kernel_win.update_connection()

        # Explorer panel
        elif self.pkey in [9, 69]:    # -> TAB/E
            self.app.explorer_win.display()
            self.resize_curses(True)  # Fix brutal resize crash



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Adapted from python daemon library written by Sander Marechal :
#
# Source : http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/www.boxedice.com
# Adaptation to Python 3.x : https://gist.github.com/andreif/cbb71b0498589dac93cb
#
# Copyright © Sander Marechal
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date :
# Last Modified : sam. 10 mars 2018 20:22:23 CET
"""
-----------
DOCSTRING

Generic linux daemon base class for python 3.x.

@author: Cyril Desjouy
"""

import sys
import os
import time
import atexit
import signal


class Daemon:
    """A generic daemon class.
    Usage: subclass the daemon class and override the run() method."""

    def __init__(self, pidfile):
        self.pidfile = pidfile

    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
            sys.exit(1)

        # decouple from parent environment
        os.chdir('/')
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)

        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as f:
            f.write(pid + '\n')

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile, 'r') as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if pid:
            message = "pidfile {0} already exist. Daemon already running?\n"
            sys.stderr.write(message.format(self.pidfile))
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""

        # Get the pid from the pidfile
        try:
            with open(self.pidfile, 'r') as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if not pid:
            message = "pidfile {0} does not exist. Daemon not running?\n"
            sys.stderr.write(message.format(self.pidfile))
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            e = str(err.args)
            if e.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err.args))
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def run(self):
        """You should override this method when you subclass Daemon.
        It will be called after the process has been daemonized by
        start() or restart()."""
